// **********************************************
//     Heap File Page Class
//     $Id: hfpage.C,v 1.15 1997/01/02 12:29:19 flisakow Exp $
// **********************************************

#include <iostream>
#include <stdlib.h>
#include <memory.h>

#include "hfpage.h"
#include "heapfile.h"
#include "buf.h"
#include "db.h"

#ifdef MULTIUSER
#    include "recovery_mgr.h"
#endif

static const char *hfpageErrMsgs[] = {
  "Invalid slot number"
};
static error_string_table hfTable( HEAPFILE, hfpageErrMsgs );

// **********************************************************
// page class constructor

void HFPage::init(PageId pageNo)
{
#ifdef MULTIUSER
    char tmp_buf[DPFIXED];
    memcpy(tmp_buf, (void *)&(slot[0]), DPFIXED);
#endif

    nextPage = prevPage = INVALID_PAGE;

    slotCnt  = 0;                // no slots in use
    curPage  = pageNo;

    usedPtr   = sizeof(data);    // offset of used space in data array

    freeSpace = sizeof(data);    // amount of space available

#ifdef MULTIUSER
    MINIBASE_RECMGR->WriteUpdateLog(DPFIXED, pageNo, sizeof(data), tmp_buf,
                (char *)&(slot[0]), (Page *)this);
#endif
}

// **********************************************************
// dump page utlity
void HFPage::dumpPage()
{
    int i;

	//    cout << "dumpPage, this: " << this << endl;
	//    cout << "curPage= " << curPage << ", nextPage=" << nextPage << endl;
    cout << "usedPtr=" << usedPtr << ",  freeSpace=" << freeSpace
         << ", slotCnt=" << slotCnt << endl;
    
    for (i=0; i < slotCnt; i++) {
        cout << "slot["<< i <<"].offset=" << slot[i].offset
             << ", slot["<< i << "].length=" << slot[i].length << endl; 
    }
}

// **********************************************************
PageId HFPage::getPrevPage()
{
    return prevPage;
}

// **********************************************************
void HFPage::setPrevPage(PageId pageNo)
{
#ifdef MULTIUSER
    MINIBASE_RECMGR->WriteUpdateLog(sizeof(PageId), 
                curPage, (char*)&prevPage - (char*)this,
                (char *)&prevPage, (char *)&pageNo,
                (Page *)this);
#endif

    prevPage = pageNo;
}

// **********************************************************
void HFPage::setNextPage(PageId pageNo)
{
#ifdef MULTIUSER
    MINIBASE_RECMGR->WriteUpdateLog(sizeof(PageId), 
                curPage, (char*)&nextPage - (char*)this,
                (char *)&nextPage, (char *)&pageNo,
                (Page *)this);
#endif

    nextPage = pageNo;
}

// **********************************************************
PageId HFPage::getNextPage()
{
    return nextPage;
}

// **********************************************************
// Add a new record to the page. Returns OK if everything went OK
// otherwise, returns DONE if sufficient space does not exist
// RID of the new record is returned via rid parameter.
Status HFPage::insertRecord(char* recPtr, int recLen, RID& rid)
{
    RID tmpRid;
    int spaceNeeded = recLen + sizeof(slot_t);
#ifdef    MULTIUSER
    Status status;
#endif

    // Start by checking if sufficient space exists.
    // This is an upper bound check. May not actually need a slot
    // if we can find an empty one.

    if (spaceNeeded > freeSpace) {
        return DONE;
    } else {

        // look for an empty slot

        int i;
        for (i=0; i < slotCnt; i++) {
            if (slot[i].length == EMPTY_SLOT)
                break;
        }

        // at this point we have either found an empty slot 
        // or i will be equal to slotCnt.  In either case,
        // we can just use i as the slot index

        // adjust free space

#ifdef MULTIUSER
        void* tmp_buf1 = new char[DPFIXED];
        memcpy(tmp_buf1, (void *) &(slot[0]), DPFIXED);
#endif    

        if (i == slotCnt) {

              // using a new slot
            freeSpace -= spaceNeeded;
            slotCnt++;

        } else {

              // reusing an existing slot 
            freeSpace -= recLen;
        }

        // use existing value of slotCnt as the index into slot array
        // use before incrementing because constructor sets the initial
        // value to 0

#ifdef MULTIUSER
        void* tmp_buf3 = new char[sizeof(slot_t)];
        memcpy(tmp_buf3, (void *) &(slot[i]), sizeof(slot_t));
#endif

        usedPtr -= recLen;    // adjust usedPtr

        slot[i].offset = usedPtr;
        slot[i].length = recLen;

#ifdef MULTIUSER
        status = MINIBASE_RECMGR->WriteUpdateLog(sizeof(slot_t), curPage, 
                 sizeof(data) +i*sizeof(slot_t) , 
                 tmp_buf3, (char *)&slot[i], (Page *)this);

        if( status != OK)
            return MINIBASE_CHAIN_ERROR(HEAPFILE,status);

        delete [] tmp_buf3;

        void* tmp_buf2 = new char[recLen];
        memcpy(tmp_buf2, &(data[usedPtr]), recLen);
#endif

        memcpy(&data[usedPtr],recPtr,recLen); // copy data onto the data page

#ifdef MULTIUSER
        status = MINIBASE_RECMGR->WriteUpdateLog(recLen, curPage, usedPtr,
                 tmp_buf2, recPtr, (Page *)this);
        if( status != OK)
            return MINIBASE_CHAIN_ERROR(HEAPFILE,status);
        delete [] tmp_buf2;
#endif

        tmpRid.pageNo = curPage;
        tmpRid.slotNo = i;
        rid = tmpRid;

#ifdef MULTIUSER
        status = MINIBASE_RECMGR->WriteUpdateLog(DPFIXED, curPage,
                 sizeof data, tmp_buf1, (char *)&slot[0], (Page *)this);
        if (status != OK)
            return MINIBASE_CHAIN_ERROR(HEAPFILE,status);
        delete [] tmp_buf1;
#endif

        return OK;
    }
}

// **********************************************************
// Delete a record from a page. Returns OK if everything went okay.
// Compacts remaining records but leaves a hole in the slot array.
// Use memmove() rather than memcpy() as space may overlap.
Status HFPage::deleteRecord(const RID& rid)
{
#ifdef MULTIUSER
    Status status;
#endif

    int slotNo = rid.slotNo;
#ifdef    MULTIUSER
    void* tmp_buf = new char[MAX_SPACE];
#endif


    // first check if the record being deleted is actually valid
    if ((slotNo >= 0) && (slotNo < slotCnt) && (slot[slotNo].length > 0)) {

#ifdef MULTIUSER
        memcpy(tmp_buf, (void *)this, MAX_SPACE);
#endif

        // valid slot

        // The records always need to be compacted, as they are
        // not necessarily stored on the page in the order that
        // they are listed in the slot index.  
        
        int offset = slot[slotNo].offset; // offset of record being deleted
        int recLen = slot[slotNo].length; // length of record being deleted

        char* newSpot = &(data[usedPtr + recLen]);

            // calculate the num of bytes to move
        int sze = offset - usedPtr;
        memmove(newSpot, &(data[usedPtr]), sze); // shift bytes to the right


        // now need to adjust offsets of all valid slots that refer 
        // to the left of the record being removed. (by the size of the hole)

        int i;
        for (i = 0; i < slotCnt; i++) {
            if ((slot[i].length >= 0)
                   && (slot[i].offset < slot[slotNo].offset))
                slot[i].offset += recLen;
        }

        usedPtr   += recLen;   // move used Ptr forward
        freeSpace += recLen;   // increase freespace by size of hole

        slot[slotNo].length = EMPTY_SLOT;  // mark slot free
        slot[slotNo].offset =  0;

#ifdef MULTIUSER
        status = MINIBASE_RECMGR->WriteUpdateLog(MAX_SPACE, curPage, 0,
                 tmp_buf, (char *)this, (Page *)this);
        if( status != OK)
            return MINIBASE_CHAIN_ERROR(HEAPFILE,status);
        delete [] tmp_buf;    
#endif
        return OK;
        
    } else {
        return MINIBASE_FIRST_ERROR( HEAPFILE, INVALID_SLOTNO );
    }
}

// **********************************************************
// returns RID of first record on page
Status HFPage::firstRecord(RID& firstRid)
{
    RID tmpRid;
    int i;

    // find the first non-empty slot

    for (i=0; i < slotCnt; i++) {
        if (slot[i].length != EMPTY_SLOT)
            break;
    }

    if ((i == slotCnt) || (slot[i].length == EMPTY_SLOT)) {
        return DONE;
    }

      // found a non-empty slot

    tmpRid.pageNo = curPage;
    tmpRid.slotNo = i;
    firstRid = tmpRid;

    return OK;
}

// **********************************************************
// returns RID of next record on the page
// returns DONE if no more records exist on the page; otherwise OK
Status HFPage::nextRecord (RID curRid, RID& nextRid)
{
    RID tmpRid;
    int i; 

      // find the next non-empty slot
    for (i=curRid.slotNo+1; i < slotCnt; i++) {
        if (slot[i].length != EMPTY_SLOT)
            break;
    }

    if ((i >= slotCnt) || (slot[i].length == EMPTY_SLOT)) {
        return DONE;
    }

      // found a non-empty slot
    tmpRid.pageNo = curPage;
    tmpRid.slotNo = i;
    nextRid = tmpRid;

    return OK;
}

// **********************************************************
// returns length and copies out record with RID rid
Status HFPage::getRecord(RID rid, char* recPtr, int& recLen)
{
    int slotNo = rid.slotNo;
    int offset;

    if ((slotNo < slotCnt) && (slot[slotNo].length > 0)) {

        offset = slot[slotNo].offset;   // extract offset in data[]

         // copy out the record

        recLen = slot[slotNo].length;   // return length of record
        memcpy(recPtr, &(data[offset]), recLen);

        return OK;
    } else {
        return MINIBASE_FIRST_ERROR( HEAPFILE, INVALID_SLOTNO );
    }
}

// **********************************************************
// returns length and pointer to record with RID rid.  The difference
// between this and getRecord is that getRecord copies out the record
// into recPtr, while this function returns a pointer to the record
// in recPtr.
Status HFPage::returnRecord(RID rid, char*& recPtr, int& recLen)
{
    int slotNo = rid.slotNo;
    int offset;

    if ((slotNo < slotCnt) && (slot[slotNo].length > 0)) {

        offset = slot[slotNo].offset;  // extract offset in data[]
        recLen = slot[slotNo].length;  // return length of record
        recPtr = &(data[offset]);      // return pointer to record

        return OK;
    } else {
        return MINIBASE_FIRST_ERROR( HEAPFILE, INVALID_SLOTNO );
    }
}

#ifdef    MULTIUSER
// **********************************************************
// Returns the offset within the data[] area - required for a
// WriteUpdateLogRec call.
Status HFPage::returnOffset(RID rid, int& offset)
{
    int slotNo = rid.slotNo;

    if ((slotNo < slotCnt) && (slot[slotNo].length > 0)) {
        offset = slot[slotNo].offset; // extract offset in data[]
        return OK;
    } else {
        return MINIBASE_FIRST_ERROR( HEAPFILE, INVALID_SLOTNO );
    }
}

#endif
// **********************************************************
// Returns the amount of available space on the heap file page.
// You will have to compare it with the size of the record to
// see if the record will fit on the page.
int HFPage::available_space(void)
{
    return freeSpace - sizeof(slot_t);
}

// **********************************************************
// Returns True if the HFPage is empty, and False otherwise.
// It scans the slot directory looking for a non-empty slot.
bool HFPage::empty(void)
{
    int i;

      // look for an empty slot
    for (i=0; i < slotCnt; i++)
        if (slot[i].length != EMPTY_SLOT)
            return false;

    return true;
}

// **********************************************************
// Start from the begining of the slot directory.
// We maintain two offsets into the directory.
//   o first free slot       (ffs) 
//   o current scan position (current_scan_posn)
// Used slots after empty slots are copied to the ffs.
// Shifting the whole array is done rather than filling
// the holes since the array may be sorted...
void HFPage::compact_slot_dir(void)
{
   int  current_scan_posn =  0;
   int  first_free_slot   = -1;   // An invalid position.
   bool move = false;             // Move a record? -- initially false

#ifdef MULTIUSER
   char old_data[DPFIXED];
   memcpy(old_data, (void*)&slot[0], DPFIXED);
#endif

   while (current_scan_posn < slotCnt) {
       if ((slot[current_scan_posn].length == EMPTY_SLOT)
                && (move == false)) {
           move = true;
           first_free_slot = current_scan_posn;
       } else if ((slot[current_scan_posn].length != EMPTY_SLOT)
                 && (move == true)) {
//         cout << "Moving " << current_scan_posn << " --> "
//              << first_free_slot << endl;
           slot[first_free_slot].length = slot[current_scan_posn].length;
           slot[first_free_slot].offset = slot[current_scan_posn].offset;
     
             // Mark the current_scan_posn as empty
           slot[current_scan_posn].length = EMPTY_SLOT;

             // Now make the first_free_slot point to the next free slot.
           first_free_slot++;

                 // slot[current_scan_posn].length == EMPTY_SLOT !!
           while (slot[first_free_slot].length != EMPTY_SLOT)  
               first_free_slot++;
       }
       current_scan_posn++;
    }

   if (move == true) {
       // Adjust amount of free space on page and slotCnt
       freeSpace += sizeof(slot_t) * (slotCnt - first_free_slot);
       slotCnt = first_free_slot;
   }

#ifdef MULTIUSER
   MINIBASE_RECMGR->WriteUpdateLog(DPFIXED, curPage,sizeof(data),
                old_data, (char*)&(slot[0]), (Page*) this);
#endif
}

// **********************************************************
